from golgi.config import configurable
from series.logging import Logging

from series.api_client import ApiClient as Base


@configurable(library_client=['rest_api_port', 'rest_api_url'])
class ApiClient(Base):
    command = Base.command

    @command('series season episode', 'Create an episode with the ' +
             'supplied metadata')
    def create_episode(self, series, season, episode):
        data = dict(episode=episode)
        path = 'series/{}/seasons/{}/episodes'.format(series, season)
        response = self.post(path, body=data)
        self._info(response)
        return response

    @command('series season episode subfps', 'Set the episode\'s subfps')
    def subfps(self, series, season, episode, subfps):
        data = dict(subfps=subfps)
        path = 'series/{}/seasons/{}/episodes/{}'.format(series, season,
                                                         episode)
        response = self.put(path, body=data)
        self._info(response)
        return response

__all__ = ['ApiClient']



import os

from golgi.config import (Config, PathConfigOption, ListConfigOption,
                          FileSizeConfigOption, FloatConfigOption,
                          BoolConfigOption, IntConfigOption)

metadata = dict(parents=['series.store_episode', 'series.subsync',
                         'series.etvdb'])


def reset_config():
    path_template = '{name}/s{season}/{name}_{season:0>2}x{episode:0>2}.{ext}'
    lib_help = 'Add new releases to the library via REST API'
    run_help = ('Comma-separated list of components to start (feed_poller, ' +
                'downloader, archiver, subsyncer, rest_api, link_handler)')
    omit_help = 'List of components not to start'
    Config.override('store_episode', match_threshold=1., overwrite=True)
    return {
        'get': dict(
            rss_urls=ListConfigOption(),
            rss_interval=15 * 60,
            download_dir=PathConfigOption(os.getcwd()),
            run=ListConfigOption([], help=run_help),
            omit=ListConfigOption([], help=omit_help),
            use_sharehosters=True,
            min_size=FileSizeConfigOption(),
            path_template=path_template,
            rest_api_port=8110,
            rest_api_host='0.0.0.0',
            prefer_hosters=ListConfigOption(
                help='hoster domains to prefer with multiple links'),
            subtitle_retry_coefficient=FloatConfigOption(
                30, help='Initial number of minutes to wait before retrying '
                'subtitle downloads'
            ),
            library=BoolConfigOption(True, help=lib_help),
            archive_exec='',
            archive_exec_args='',
            auto_upgrade_db=True,
            link_check_procs=5,
            link_check_proc_timeout=IntConfigOption(
                30, help='timeout for link checking subprocesses in seconds'),
            sync_link_check=False,
            link_check_retry_coefficient=FloatConfigOption(
                5, help='Initial number of minutes to wait before retrying '
                'link checks'
            ),
            db_path=PathConfigOption(),
            sub_exclude=ListConfigOption([]),
            torrent_recheck_interval=3600,
            only_torrent=True,
            full_hd=True,
            min_seeders=0,
        ),
        'show_planner': dict(
            check_interval=3600 * 24,
            show_db='etvdb'
        ),
        'get_client': dict(
            rest_api_url='http://localhost', rest_api_port=8110
        )
    }



from datetime import datetime

from tvrage import api as tv
from tvrage.exceptions import (BaseError as TvRageBaseError,
                               NoNewEpisodesAnnounced)
from tvrage.util import TvrageError

from series.logging import Logging


def date_to_datetime(date):
    return datetime.fromordinal(date.toordinal())


class TVRageFacade(Logging):

    def show(self, name, sid=None):
        try:
            return tv.Show(name, sid=sid)
        except (TvrageError, TvRageBaseError) as e:
            self.log.error('Error querying tvrage show: {}'.format(e))

    def next_episode_date(self, show):
        try:
            d = show.next_episode.airdate
            return date_to_datetime(d)
        except NoNewEpisodesAnnounced:
            self.log.info(
                'No new episodes announced for "{}"'.format(show.name))
        except (TvrageError, TvRageBaseError) as e:
            self.log.error('Error querying tvrage episode date: {}'.format(e))

    def next_episode_enum(self, show):
        try:
            return (show.next_episode.season, show.next_episode.number)
        except (TvrageError, TvRageBaseError) as e:
            self.log.error('Error querying tvrage episode enum: {}'.format(e))

    def airdate(self, show, season, episode):
        rage = self.show(show.name, show.rage_id)
        if rage:
            season = rage.episodes.get(season)
            if season:
                epi = season.get(episode)
                if epi:
                    return date_to_datetime(epi.airdate)

    def season(show: tv.Show, season):
        return show.episodes[season].values()

__all__ = ['TVRageFacade']



import re
import itertools
from datetime import datetime

import requests

import lxml

from series.get.handler import ReleaseHandler, R
from series.get.model.release import ReleaseMonitor
from series.condition import LambdaCondition

from golgi.config import configurable

from tek_utils.sharehoster.torrent import SearchResultFactory, SearchResult
from tek_utils.sharehoster.kickass import NoResultsError

from amino import List, LazyList, _, F, __
from amino.lazy import lazy


class SearchQuery:

    def __init__(self, monitor: ReleaseMonitor, res: str) -> None:
        self.monitor = monitor
        self.release = self.monitor.release
        self.res = res

    @property
    def _enum(self):
        return 's{:0>2}e{:0>2}'.format(self.release.season,
                                       self.release.episode)

    @property
    def _name(self):
        return self.release.name.replace('_', ' ').replace('\'', '')

    @property
    def query(self):
        return '{} {} {}'.format(
            self._name,
            self._enum,
            self.res,
        )

    @property
    def valid(self):
        return True

    @property
    def search_string(self):
        return self.release.search_string_with_res(self.res, False)

    @lazy
    def search_re(self):
        return re.compile(self.search_string, re.I)

    @property
    def desc(self):
        return 'torrent {} {}'.format(self.release, self.res)


class DateQuery(SearchQuery):

    @property
    def _enum(self):
        return self.release.airdate.strftime('%Y-%m-%d')

    @property
    def valid(self):
        return self.release.has_airdate

    @property
    def search_string(self):
        return self.release.search_string_with_res(self.res, True)


@configurable(torrent=['search_engine'], get=['torrent_recheck_interval',
                                              'min_seeders'])
class TorrentFinder(ReleaseHandler):

    def __init__(self, releases, *a, **kw):
        super().__init__(releases, 5, 'torrent finder',
                         cooldown=self._torrent_recheck_interval, **kw)
        self._search = (self._search_tpb if self._search_engine == 'piratebay'
                        else self._search_kickass)
        self._limit = 10

    def _queries(self, monitor):
        q = lambda r: List(SearchQuery(monitor, r), DateQuery(monitor, r))
        return monitor.resolutions // q

    def _handle(self, monitor):
        self.log.debug(
            'Searching for torrent for "{}"'.format(monitor.release))
        self._update(monitor, last_torrent_search=datetime.now())
        return LazyList(self._queries(monitor))\
            .filter(_.valid)\
            .find(self._handle_query)

    def _handle_query(self, query):
        self.log.debug('Search {}: {}'.format(query.desc, query.query))
        try:
            results = List.wrap(self._search(query.query))
        except NoResultsError as e:
            self.log.debug('Error searching for torrent: {}'.format(e))
        except requests.RequestException as e:
            self.log.warn(
                'Connection failure in {} search'.format(self._search_engine))
        except lxml.etree.XMLSyntaxError as e:
            self.log.warn('Parse error in kickass results: {}'.format(e))
        else:
            return self._process_results(query, results)

    def _process_results(self, query: SearchQuery, results: List):
        return (
            self._choose_result(query, results)
            .map(F(self._add_link, query))
            .replace(True)
            .get_or_else(F(self._no_result, query, results))
        )

    def _choose_result(self, query: SearchQuery, results: List[SearchResult]):
        matcher = query.search_re
        valid = results.filter(_.seeders > self._min_seeders)
        return (
            valid
            .filter(lambda a: matcher.search(a.title))
            .filter(_.magnet_link)
            .filter_not(lambda a: query.monitor.contains_link(a.magnet_link))
            .head
        )

    def _add_link(self, query: SearchQuery, result: SearchResult):
        self.log.info('Added torrent to release {}: {} ({} seeders)'
                      .format(query.release, result.title, result.seeders))
        self._releases.add_link_by_id(query.monitor.id, result.magnet_link)

    def _no_result(self, query, results):
        self.log.debug('None of the results match the release.')
        self.log.debug('min_seeders: {}'.format(self._min_seeders))
        self.log.debug('Search string: {}'.format(query.search_string))
        self.log.debug('\n'.join([r.title for r in results]))

    def _search_tpb(self, query):
        from tek_utils.sharehoster import piratebay
        bay = piratebay.Search(query)
        return bay.run[:self._limit] / SearchResultFactory.from_tpb

    def _search_kickass(self, query):
        from tek_utils.sharehoster import kickass
        search = kickass.Search(query).order(kickass.ORDER.SEED,
                                             kickass.ORDER.DESC)
        return [SearchResultFactory.from_kickass(res) for res in
                itertools.islice(search, self._limit)]

    @property
    def _conditions(self):
        return (
            ~R('downloaded') & ~R('has_cachable_torrents') &
            LambdaCondition('recheck interval',
                            __.can_recheck(self._torrent_recheck_interval))
        )

    def activate_id(self, id):
        super().activate_id(id)
        self._releases.update_by_id(id, last_torrent_search_stamp=0)

__all__ = ['TorrentFinder']



from tek.errors import TException

class NoSuchSeriesMapping(TException):
    def __init__(self, name):
        text = 'No url mapping for series {}'.format(name)
        super().__init__(text)

class NoSubsForEpisode(TException):
    def __init__(self, series, season, episode):
        text = 'No subs for episode {} of series "{}" season {}'
        text = text.format(episode, series, season)
        super().__init__(text)



import functools

from series.db import Database, Column, Integer, Float

from tek.tools import unix_to_datetime


@functools.total_ordering
@Database.many_to_one('Episode', backref_name='watch_history')
class WatchEvent(object, metaclass=Database.DefaultMeta):
    time_begin = Column(Integer)
    time_end = Column(Integer)
    stopped_at = Column(Float)

    @property
    def begin(self):
        return unix_to_datetime(self.time_begin)

    @property
    def end(self):
        return unix_to_datetime(self.time_end)

    def __str__(self):
        templ = 'Watched {} {}x{} from {} until {}'
        return templ.format(self.episode.series.name,
                            self.episode.season.number, self.episode.number,
                            self.begin, self.end)

    def __repr__(self):
        return '<WatchEvent {}>'.format(self.time_end)

    def __lt__(self, other):
        if isinstance(other, WatchEvent):
            return self.time_end < other.time_end

    def __eq__(self, other):
        if isinstance(other, WatchEvent):
            return self.time_end == other.time_end


__all__ = ['WatchEvent']



from datetime import datetime, timedelta
from typing import Callable

from series.get.handler import ShowHandler, S
from series.get.tvdb import Tvdb

from series.condition import HandlerCondition, DynCondition, SimpleCondition
from series.get.model.show import Show
from series.get import ReleaseMonitor


class AirsToday(HandlerCondition):

    def ev(self, show):
        return (show.has_next_episode and
                show.next_episode_date < datetime.now() + timedelta(days=1))

    def describe(self, show, target):
        match = self.ev(show)
        good = match == target
        today = (show.next_episode_date > datetime.now()) and match
        desc = ('today' if today else
                show.next_episode_date.strftime('%F') if
                show.next_episode_stamp > 0 else
                'no date')
        return 'airs today[{}]'.format(self._paint(desc, good))


class CanCatchUp(SimpleCondition):

    def __init__(self, latest) -> None:
        self._latest = latest

    def current(self, show):
        return show.current_episode

    def latest(self, show):
        return self._latest(show).episode

    def ev(self, show):
        return self.current(show) > self.latest(show)

    @property
    def _desc(self):
        return 'latest > watched'

    def _repr(self, show, match):
        op = '>' if match else '<'
        return '{} {} {}'.format(self.current(show), op, self.latest(show))


class CanSchedule(DynCondition):

    def __init__(self, latest: Callable[[Show], ReleaseMonitor]) -> None:
        self._latest = latest

    def _dyn_sub(self, show):
        if self._latest(show) is None:
            return (S('has_next_episode') & AirsToday())
        else:
            return CanCatchUp(self._latest)

    def describe(self, show, target):
        s = super().describe(show, target)
        pre = 'fresh' if self._latest(show) is None else 'continuing'
        return '{} show => {}'.format(pre, s)


class ShowScheduler(ShowHandler, Tvdb):

    def __init__(self, releases, shows, **kw):
        super().__init__(shows, 5, 'show scheduler', cooldown=3600, **kw)
        self._releases = releases

    def _handle(self, show):
        latest = self._latest(show)
        if latest is None:
            self._schedule(show, show.season, show.next_episode)
        else:
            for episode in range(latest.episode + 1, show.current_episode + 1):
                self._schedule(show, show.current_season, episode)

    @property
    def _conditions(self):
        return CanSchedule(self._latest)

    def _latest(self, show):
        return self._releases.latest_for_season(show.canonical_name,
                                                show.current_season)

    def _schedule(self, show, season, episode):
        airdate = self.tvdb.airdate(show, season, episode)
        msg = 'Scheduling release "{} {}x{}" on {}'.format(
            show.name,
            season,
            episode,
            airdate
        )
        self.log.info(msg)
        self._releases.create(show.canonical_name, season, episode, airdate)
        self._commit()

__all__ = ['ShowScheduler']



class NullService(object):

    def __init__(self, *a, **kw):
        pass

    def start(self):
        pass

    def stop(self):
        pass

    def join(self):
        pass

    def is_alive(self):
        return False

__all__ = []



import requests

from flask import json

from series.logging import Logging

from series.client.errors import SeriesClientException


def command_decorator():
    doc = {}

    def command(args, description):
        def decor(func):
            doc[func.__name__] = args, description
            return func
        return decor
    return doc, command


class ApiClient(Logging):

    doc, command = command_decorator()

    def __init__(self, info_output=True):
        self._info_output = info_output

    def _url(self, path):
        return '{}:{}/{}'.format(self._rest_api_url, self._rest_api_port, path)

    def _request(self, req_type, path, body):
        headers = {'content-type': 'application/json'}
        requester = getattr(requests, req_type)
        url = self._url(path)
        self.log.debug('api request: {}'.format(url))
        try:
            response = requester(url, data=json.dumps(body), headers=headers)
        except requests.RequestException as e:
            msg = 'Request failed! ({})'.format(e)
            raise SeriesClientException(msg) from e
        else:
            if response.status_code >= 400:
                self.log.error(
                    'API response status {}'.format(response.status_code))
            try:
                _json = response.json()
            except ValueError as e:
                msg = 'Error in API request (no JSON in response)!'
                raise SeriesClientException(msg) from e
            else:
                data = _json.get('response', {})
                if isinstance(data, dict) and 'error' in data:
                    self.log.error(data['error'])
                return data

    def get(self, path, body={}):
        return self._request('get', path, body)

    def post(self, path, body={}):
        return self._request('post', path, body)

    def put(self, path, body={}):
        return self._request('put', path, body)

    def delete(self, path, body={}):
        return self._request('delete', path, body)

    def _info(self, msg):
        if self._info_output:
            self.log.info(msg)

    @command('', 'Display this help text')
    def help(self):
        if self.doc:
            maxlen = len(max(self.doc.keys(), key=len))
            pad = lambda s: s.ljust(maxlen)
            self.log.info('Available seriesd commands:')
            for name, (args, description) in self.doc.items():
                self.log.info('')
                self.log.info('{}    {}'.format(pad(name), args))
                self.log.info('  {}'.format(description))

__all__ = ['ApiClient']



import abc
import time
from threading import Thread
from datetime import datetime, timedelta

import sqlalchemy

from series.logging import Logging

from amino import List

from series.errors import SeriesDException
from series.condition import HandlerCondition, StrictCondition


class Handler(Thread, Logging, metaclass=abc.ABCMeta):

    def __init__(self, interval, description, cooldown=0):
        self._interval = interval
        self._description = description
        super().__init__(name=self._description)
        self._running = True
        self._initial_wait = 5
        self._failures = 0
        self._cooldown = cooldown
        self._last_check = {}

    def run(self):
        self.log.info('Starting {}.'.format(self._description))
        try:
            self._sanity_check()
        except SeriesDException as e:
            self.log.error('Shutting down {} with reason: {}'.format(
                self._description, e))
        else:
            self._main_loop()

    def _main_loop(self):
        next_check = datetime.now() + timedelta(seconds=self._initial_wait)
        while self._running:
            if datetime.now() > next_check:
                self._check_error_wrapper()
                next_check = self._next_check
            time.sleep(1)

    def stop(self):
        self._running = False

    def _check_error_wrapper(self):
        success = False
        try:
            success = self._check()
        except sqlalchemy.exc.InvalidRequestError as e:
            self.log.debug(e)
        except Exception as e:
            self._failure(e)
        else:
            if success:
                self._success()

    def _check(self):
        with self._lock:
            return self._current.effect(self._process).is_just

    def _process(self, item):
        if hasattr(item, 'id'):
            self._last_check[str(item.id)] = time.time()
        self._handle(item)

    @property
    def _next_check(self):
        return datetime.now() + timedelta(seconds=self._interval)

    def _handle(self, task):
        pass

    @property
    def _current(self):
        return self._choose(self._qualified)

    def _choose(self, qualified):
        return qualified.head

    def _cool(self, item):
        return (not hasattr(item, 'id') or
                self._time_since_last(item) > self._cooldown)

    def _time_since_last(self, item):
        return time.time() - self._last_check.get(str(item.id), 0)

    def activate_id(self, id):
        if str(id) in self._last_check:
            del self._last_check[str(id)]

    @abc.abstractproperty
    def _candidates(self):
        ...

    def _qualify(self, candidate):
        return self._conditions.ev(candidate)

    @property
    def _qualified(self):
        return List.wrap(self._candidates)\
            .filter(self._cool)\
            .filter(self._qualify)

    def explain(self, item):
        cond = self._conditions.describe(item, target=True)
        return dict(name=self._description, cond=cond)

    @property
    def _conditions(self) -> HandlerCondition:
        return StrictCondition(False)

    def _commit(self):
        pass

    def _sanity_check(self):
        pass

    def _failure(self, exc):
        self._failures += 1
        import traceback
        self.log.error('Error in {}:'.format(self._description))
        self.log.error(exc)
        traceback.print_exc()
        if self._failures >= 20:
            msg = '20 failures in a row in {}, shutting down.'
            self.log.error(msg.format(self._description))
            self.stop()

    def _success(self):
        self._failures = 0

    @property
    def _lock(self):
        raise NotImplementedError('Handler._lock')

__all__ = ['Handler']



"""add TorrentLink.dead

Revision ID: 566cc0cdd00a
Revises: 376061fc33e
Create Date: 2016-05-22 18:53:55.675715

"""

# revision identifiers, used by Alembic.
revision = '566cc0cdd00a'
down_revision = '376061fc33e'

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.add_column('torrent_link', sa.Column('dead', sa.Boolean(), nullable=True))
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.add_column('release_monitor', sa.Column('torrent_cached', sa.BOOLEAN(), nullable=True))
    ### end Alembic commands ###



import re
from datetime import datetime

import requests

from sqlalchemy import Column, String, Integer, Boolean

from series.logging import Logging
from golgi.config import configurable
from tek.tools import unix_to_datetime, datetime_to_unix, sizeof_fmt
from tek.util.decorator import lazy_property

from tek_utils.sharehoster import downloader
from tek_utils.sharehoster.errors import ShareHosterError
from tek_utils.sharehoster.torrent import is_torrent

from amino import Map

from series.util import domain
from series.db import Database

from series.get.model.torrent import TorrentProxy


@configurable(get=['min_size'])
class LinkChecker(Logging):

    def __init__(self, link):
        self.link = link
        self.done = 0
        self.status = 0
        self.dead = 0
        self.reason = ''
        self.size = 0
        self.multipart = 0

    @property
    def result(self):
        dl = self.downloader
        if dl:
            status = dl.status
            if status.success:
                if status.unknown:
                    self.status = HttpLink.CHECK_FAILED
                else:
                    self.status = HttpLink.CHECKED
                self.dead = status.error
            else:
                self.status = HttpLink.CHECK_FAILED
            if self.dead:
                self.reason = 'down'
            else:
                self.size = dl.file_size
                if self._min_size > 0 and self.size > self._min_size:
                    self.status = HttpLink.CHECKED
                self.multipart = bool(re.search('part\d+\.rar$',
                                                str(dl.file_path) or ''))
            dl.close()
        return self.status, self.dead, self.reason, self.size, self.multipart

    @property
    def downloader(self):
        try:
            return downloader(self.link.url)
        except (ShareHosterError, requests.RequestException) as e:
            self.status = HttpLink.CHECK_FAILED
            self.reason = str(e)
            self.log.debug('LinkChecker: {}'.format(e))


@Database.many_to_one('ReleaseMonitor', ref_name='monitor')
class Link(object, metaclass=Database.DefaultMeta):
    url = Column(String)
    failures = Column(Integer)

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self.failures = 0

    @classmethod
    def create(self, url):
        return TorrentLink(url) if is_torrent(url) else HttpLink(url)

    @property
    def invalid(self):
        return not self.valid

    @property
    def valid(self):
        return False


@configurable(get=['min_size', 'sync_link_check',
                   'link_check_retry_coefficient'])
class HttpLink(Link):
    UNCHECKED = 1
    CHECK_FAILED = 2
    CHECKED = 3

    status = Column(Integer)
    dead = Column(Boolean)
    multipart = Column(Boolean)
    time_checked = Column(Integer)
    checking = Column(Boolean)
    size = Column(Integer)
    reason = Column(String)

    def __init__(self, url):
        super().__init__(url=url)
        self.status = self.UNCHECKED
        self.dead = False
        self.multipart = False
        self.checking = False
        self.size = -1
        self.reason = 'unknown'

    @property
    def valid(self):
        if self.url is not None:
            if self.unchecked and self._sync_link_check:
                self.check()
            return self.checked and not (self.dead or self.too_small or
                                         self.multipart)

    @property
    def potential(self):
        if self.url is not None:
            return self.valid or (self.check_failed and not self.too_small and
                                  not self.multipart)

    @property
    def too_small(self):
        return self.size > 0 and self.size < self._min_size

    def check(self):
        self.last_check = datetime.now()
        self.set_status(*LinkChecker(self).result)

    @property
    def status_str(self):
        if self.invalid and self.potential:
            status = 'Unknown link'
        else:
            pre = 'Inv' if self.invalid else 'V'
            status = '{}alid link'.format(pre)
        reason = ': {}'.format(self._reason) if self._reason else ''
        return '{}{} ({})'.format(status, reason, self.url)

    @property
    def _reason(self):
        reason = ''
        if self.dead:
            reason = 'dead ({})'.format(self.reason)
        elif self.too_small:
            reason = 'too small ({})'.format(self.size_str)
        elif self.multipart:
            reason = 'multipart archive'
        elif self.failed:
            reason = 'check failed ({})'.format(self.reason)
        elif self.unchecked:
            reason = 'unchecked'
        else:
            reason = self.reason
        return reason

    @property
    def domain(self):
        return domain(self.url) if self.url else ''

    @property
    def size_str(self):
        return sizeof_fmt(self.size)

    @property
    def last_check(self):
        if self.time_checked:
            return unix_to_datetime(self.time_checked)

    @last_check.setter
    def last_check(self, moment):
        self.time_checked = datetime_to_unix(moment)

    @property
    def checked(self):
        return self.status == self.CHECKED

    @property
    def checkable(self):
        return not self.checking and self.status == self.UNCHECKED

    @property
    def recheckable(self):
        return (not self.checking and self.status == self.CHECK_FAILED and
                self.retry_check)

    @property
    def retry_check(self):
        return (
            self.status != self.CHECKED and
            (self.failures == 0 or
             self.last_check and self._time_since_last_check >
             self._next_retry)
        )

    @property
    def _time_since_last_check(self):
        return (datetime.now() - self.last_check).total_seconds()

    @property
    def _next_retry(self):
        return (self._link_check_retry_coefficient * 60 *
                (2 ** (self.failures - 1)))

    def __str__(self):
        s = '<Link {} status: {} dead: {}'.format(
            self.url, self.status, self.dead
        )
        if self.reason:
            s += ' reason: {}'.format(self.reason)
        return s + '>'

    def __repr__(self):
        return str(self)

    def set_status(self, status, dead, reason, size, multipart):
        self.status = status
        self.dead = bool(dead)
        self.reason = reason
        self.size = int(size)
        self.multipart = bool(multipart)
        if status == self.CHECK_FAILED:
            self.check_failed()

    def check_failed(self):
        self.status = HttpLink.CHECK_FAILED
        self.failures += 1

    @property
    def failed(self):
        return self.status == HttpLink.CHECK_FAILED

    @property
    def unchecked(self):
        return self.status == HttpLink.UNCHECKED

    @property
    def download_url(self):
        return self.url

    @property
    def info(self):
        return Map(
            url=self.url,
            status=self.status,
            dead=self.dead,
            multipart=self.multipart,
            time_checked=self.time_checked,
            checking=self.checking,
            size=self.size,
            reason=self.reason,
        )


class TorrentLink(Link):
    cached = Column(Boolean)
    dead = Column(Boolean)

    def __init__(self, url):
        super().__init__(url=url)
        self.cached = False
        self.dead = False

    @property
    def download_url(self):
        return self.torrent.download_url

    @lazy_property
    def torrent(self):
        return TorrentProxy(self)

    @property
    def valid(self):
        return (self.url is not None and
                self.cached and
                (not self.dead) and
                self.download_url is not None)

    @property
    def domain(self):
        return self.torrent.domain

    def check_failed(self):
        self.cached = False
        self.failures += 1

    def __str__(self):
        s = '<TorrentLink valid: {} dead: {} cachable: {} url: {}>'
        return s.format(self.valid, self.dead, self.cachable, self.url)

    @property
    def status_str(self):
        pre = 'Inv' if self.invalid else 'V'
        status = '{}alid torrent'.format(pre)
        return '{} ({})'.format(status, self.url.split('&')[0])

    @property
    def caching(self):
        return self.torrent.caching

    @property
    def cachable(self):
        return not (self.caching or self.valid or self.dead)

    @property
    def info(self):
        return Map(
            url=self.url,
            cached = self.cached,
            dead = self.dead,
        )

__all__ = ['Link', 'TorrentLink']



from series.condition import AttrCondition
from series.handler import Handler
from series.get.model.release import ReleaseMonitor
from series.get.model.link import Link
from series.get.model.show import Show


class ReleaseAttr(AttrCondition[ReleaseMonitor]):

    def __init__(self, attr: str) -> None:
        super().__init__('release', attr)

R = ReleaseAttr


class LinkAttr(AttrCondition[Link]):

    def __init__(self, attr: str, target=True) -> None:
        super().__init__('link', attr, target)


L = LinkAttr


class ShowAttr(AttrCondition[Show]):

    def __init__(self, attr: str, target=True) -> None:
        super().__init__('show', attr, target)


S = ShowAttr


class BaseHandler(Handler):

    def __init__(self, data, interval, description, **kw):
        self._data = data
        super().__init__(interval, description, **kw)

    @property
    def _candidates(self):
        return self._data.all

    @property
    def _lock(self):
        return self._data.lock


class ReleaseHandler(BaseHandler):

    @property
    def _releases(self):
        return self._data

    def _update(self, monitor, **data):
        self._releases.update_by_id(monitor.id, **data)


class ShowHandler(BaseHandler):

    @property
    def _shows(self):
        return self._data

__all__ = ['ReleaseHandler', 'ShowHandler']
